# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.8
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.
from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_cvxcore')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_cvxcore')
    _cvxcore = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_cvxcore', [dirname(__file__)])
        except ImportError:
            import _cvxcore
            return _cvxcore
        try:
            _mod = imp.load_module('_cvxcore', fp, pathname, description)
        finally:
            if fp is not None:
                fp.close()
        return _mod
    _cvxcore = swig_import_helper()
    del swig_import_helper
else:
    import _cvxcore
del _swig_python_version_info

try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0


class SwigPyIterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, SwigPyIterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, SwigPyIterator, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _cvxcore.delete_SwigPyIterator
    __del__ = lambda self: None

    def value(self):
        return _cvxcore.SwigPyIterator_value(self)

    def incr(self, n=1):
        return _cvxcore.SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        return _cvxcore.SwigPyIterator_decr(self, n)

    def distance(self, x):
        return _cvxcore.SwigPyIterator_distance(self, x)

    def equal(self, x):
        return _cvxcore.SwigPyIterator_equal(self, x)

    def copy(self):
        return _cvxcore.SwigPyIterator_copy(self)

    def __next__(self):
        return _cvxcore.SwigPyIterator_next(self)

    def __next__(self):
        return _cvxcore.SwigPyIterator___next__(self)

    def previous(self):
        return _cvxcore.SwigPyIterator_previous(self)

    def advance(self, n):
        return _cvxcore.SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        return _cvxcore.SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        return _cvxcore.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        return _cvxcore.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        return _cvxcore.SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        return _cvxcore.SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        return _cvxcore.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self
SwigPyIterator_swigregister = _cvxcore.SwigPyIterator_swigregister
SwigPyIterator_swigregister(SwigPyIterator)


_cvxcore.VARIABLE_swigconstant(_cvxcore)
VARIABLE = _cvxcore.VARIABLE

_cvxcore.PROMOTE_swigconstant(_cvxcore)
PROMOTE = _cvxcore.PROMOTE

_cvxcore.MUL_swigconstant(_cvxcore)
MUL = _cvxcore.MUL

_cvxcore.RMUL_swigconstant(_cvxcore)
RMUL = _cvxcore.RMUL

_cvxcore.MUL_ELEM_swigconstant(_cvxcore)
MUL_ELEM = _cvxcore.MUL_ELEM

_cvxcore.DIV_swigconstant(_cvxcore)
DIV = _cvxcore.DIV

_cvxcore.SUM_swigconstant(_cvxcore)
SUM = _cvxcore.SUM

_cvxcore.NEG_swigconstant(_cvxcore)
NEG = _cvxcore.NEG

_cvxcore.INDEX_swigconstant(_cvxcore)
INDEX = _cvxcore.INDEX

_cvxcore.TRANSPOSE_swigconstant(_cvxcore)
TRANSPOSE = _cvxcore.TRANSPOSE

_cvxcore.SUM_ENTRIES_swigconstant(_cvxcore)
SUM_ENTRIES = _cvxcore.SUM_ENTRIES

_cvxcore.TRACE_swigconstant(_cvxcore)
TRACE = _cvxcore.TRACE

_cvxcore.RESHAPE_swigconstant(_cvxcore)
RESHAPE = _cvxcore.RESHAPE

_cvxcore.DIAG_VEC_swigconstant(_cvxcore)
DIAG_VEC = _cvxcore.DIAG_VEC

_cvxcore.DIAG_MAT_swigconstant(_cvxcore)
DIAG_MAT = _cvxcore.DIAG_MAT

_cvxcore.UPPER_TRI_swigconstant(_cvxcore)
UPPER_TRI = _cvxcore.UPPER_TRI

_cvxcore.CONV_swigconstant(_cvxcore)
CONV = _cvxcore.CONV

_cvxcore.HSTACK_swigconstant(_cvxcore)
HSTACK = _cvxcore.HSTACK

_cvxcore.VSTACK_swigconstant(_cvxcore)
VSTACK = _cvxcore.VSTACK

_cvxcore.SCALAR_CONST_swigconstant(_cvxcore)
SCALAR_CONST = _cvxcore.SCALAR_CONST

_cvxcore.DENSE_CONST_swigconstant(_cvxcore)
DENSE_CONST = _cvxcore.DENSE_CONST

_cvxcore.SPARSE_CONST_swigconstant(_cvxcore)
SPARSE_CONST = _cvxcore.SPARSE_CONST

_cvxcore.NO_OP_swigconstant(_cvxcore)
NO_OP = _cvxcore.NO_OP

_cvxcore.KRON_swigconstant(_cvxcore)
KRON = _cvxcore.KRON
class LinOp(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, LinOp, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, LinOp, name)
    __repr__ = _swig_repr
    __swig_setmethods__["type"] = _cvxcore.LinOp_type_set
    __swig_getmethods__["type"] = _cvxcore.LinOp_type_get
    if _newclass:
        type = _swig_property(_cvxcore.LinOp_type_get, _cvxcore.LinOp_type_set)
    __swig_setmethods__["size"] = _cvxcore.LinOp_size_set
    __swig_getmethods__["size"] = _cvxcore.LinOp_size_get
    if _newclass:
        size = _swig_property(_cvxcore.LinOp_size_get, _cvxcore.LinOp_size_set)
    __swig_setmethods__["args"] = _cvxcore.LinOp_args_set
    __swig_getmethods__["args"] = _cvxcore.LinOp_args_get
    if _newclass:
        args = _swig_property(_cvxcore.LinOp_args_get, _cvxcore.LinOp_args_set)
    __swig_setmethods__["data_ndim"] = _cvxcore.LinOp_data_ndim_set
    __swig_getmethods__["data_ndim"] = _cvxcore.LinOp_data_ndim_get
    if _newclass:
        data_ndim = _swig_property(_cvxcore.LinOp_data_ndim_get, _cvxcore.LinOp_data_ndim_set)
    __swig_setmethods__["sparse"] = _cvxcore.LinOp_sparse_set
    __swig_getmethods__["sparse"] = _cvxcore.LinOp_sparse_get
    if _newclass:
        sparse = _swig_property(_cvxcore.LinOp_sparse_get, _cvxcore.LinOp_sparse_set)
    __swig_setmethods__["sparse_data"] = _cvxcore.LinOp_sparse_data_set
    __swig_getmethods__["sparse_data"] = _cvxcore.LinOp_sparse_data_get
    if _newclass:
        sparse_data = _swig_property(_cvxcore.LinOp_sparse_data_get, _cvxcore.LinOp_sparse_data_set)
    __swig_setmethods__["dense_data"] = _cvxcore.LinOp_dense_data_set
    __swig_getmethods__["dense_data"] = _cvxcore.LinOp_dense_data_get
    if _newclass:
        dense_data = _swig_property(_cvxcore.LinOp_dense_data_get, _cvxcore.LinOp_dense_data_set)
    __swig_setmethods__["slice"] = _cvxcore.LinOp_slice_set
    __swig_getmethods__["slice"] = _cvxcore.LinOp_slice_get
    if _newclass:
        slice = _swig_property(_cvxcore.LinOp_slice_get, _cvxcore.LinOp_slice_set)

    def __init__(self):
        this = _cvxcore.new_LinOp()
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def has_constant_type(self):
        return _cvxcore.LinOp_has_constant_type(self)

    def set_dense_data(self, matrix):
        return _cvxcore.LinOp_set_dense_data(self, matrix)

    def set_sparse_data(self, data, row_idxs, col_idxs, rows, cols):
        return _cvxcore.LinOp_set_sparse_data(self, data, row_idxs, col_idxs, rows, cols)
    __swig_destroy__ = _cvxcore.delete_LinOp
    __del__ = lambda self: None
LinOp_swigregister = _cvxcore.LinOp_swigregister
LinOp_swigregister(LinOp)
cvar = _cvxcore.cvar
CONSTANT_ID = cvar.CONSTANT_ID

class ProblemData(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, ProblemData, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, ProblemData, name)
    __repr__ = _swig_repr
    __swig_setmethods__["V"] = _cvxcore.ProblemData_V_set
    __swig_getmethods__["V"] = _cvxcore.ProblemData_V_get
    if _newclass:
        V = _swig_property(_cvxcore.ProblemData_V_get, _cvxcore.ProblemData_V_set)
    __swig_setmethods__["I"] = _cvxcore.ProblemData_I_set
    __swig_getmethods__["I"] = _cvxcore.ProblemData_I_get
    if _newclass:
        I = _swig_property(_cvxcore.ProblemData_I_get, _cvxcore.ProblemData_I_set)
    __swig_setmethods__["J"] = _cvxcore.ProblemData_J_set
    __swig_getmethods__["J"] = _cvxcore.ProblemData_J_get
    if _newclass:
        J = _swig_property(_cvxcore.ProblemData_J_get, _cvxcore.ProblemData_J_set)
    __swig_setmethods__["const_vec"] = _cvxcore.ProblemData_const_vec_set
    __swig_getmethods__["const_vec"] = _cvxcore.ProblemData_const_vec_get
    if _newclass:
        const_vec = _swig_property(_cvxcore.ProblemData_const_vec_get, _cvxcore.ProblemData_const_vec_set)
    __swig_setmethods__["id_to_col"] = _cvxcore.ProblemData_id_to_col_set
    __swig_getmethods__["id_to_col"] = _cvxcore.ProblemData_id_to_col_get
    if _newclass:
        id_to_col = _swig_property(_cvxcore.ProblemData_id_to_col_get, _cvxcore.ProblemData_id_to_col_set)
    __swig_setmethods__["const_to_row"] = _cvxcore.ProblemData_const_to_row_set
    __swig_getmethods__["const_to_row"] = _cvxcore.ProblemData_const_to_row_get
    if _newclass:
        const_to_row = _swig_property(_cvxcore.ProblemData_const_to_row_get, _cvxcore.ProblemData_const_to_row_set)

    def getV(self, values):
        return _cvxcore.ProblemData_getV(self, values)

    def getI(self, values):
        return _cvxcore.ProblemData_getI(self, values)

    def getJ(self, values):
        return _cvxcore.ProblemData_getJ(self, values)

    def getConstVec(self, values):
        return _cvxcore.ProblemData_getConstVec(self, values)

    def __init__(self):
        this = _cvxcore.new_ProblemData()
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _cvxcore.delete_ProblemData
    __del__ = lambda self: None
ProblemData_swigregister = _cvxcore.ProblemData_swigregister
ProblemData_swigregister(ProblemData)

class IntVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, IntVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, IntVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _cvxcore.IntVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __bool__(self):
        return _cvxcore.IntVector___nonzero__(self)

    def __bool__(self):
        return _cvxcore.IntVector___bool__(self)

    def __len__(self):
        return _cvxcore.IntVector___len__(self)

    def __getslice__(self, i, j):
        return _cvxcore.IntVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _cvxcore.IntVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _cvxcore.IntVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _cvxcore.IntVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _cvxcore.IntVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _cvxcore.IntVector___setitem__(self, *args)

    def pop(self):
        return _cvxcore.IntVector_pop(self)

    def append(self, x):
        return _cvxcore.IntVector_append(self, x)

    def empty(self):
        return _cvxcore.IntVector_empty(self)

    def size(self):
        return _cvxcore.IntVector_size(self)

    def swap(self, v):
        return _cvxcore.IntVector_swap(self, v)

    def begin(self):
        return _cvxcore.IntVector_begin(self)

    def end(self):
        return _cvxcore.IntVector_end(self)

    def rbegin(self):
        return _cvxcore.IntVector_rbegin(self)

    def rend(self):
        return _cvxcore.IntVector_rend(self)

    def clear(self):
        return _cvxcore.IntVector_clear(self)

    def get_allocator(self):
        return _cvxcore.IntVector_get_allocator(self)

    def pop_back(self):
        return _cvxcore.IntVector_pop_back(self)

    def erase(self, *args):
        return _cvxcore.IntVector_erase(self, *args)

    def __init__(self, *args):
        this = _cvxcore.new_IntVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _cvxcore.IntVector_push_back(self, x)

    def front(self):
        return _cvxcore.IntVector_front(self)

    def back(self):
        return _cvxcore.IntVector_back(self)

    def assign(self, n, x):
        return _cvxcore.IntVector_assign(self, n, x)

    def resize(self, *args):
        return _cvxcore.IntVector_resize(self, *args)

    def insert(self, *args):
        return _cvxcore.IntVector_insert(self, *args)

    def reserve(self, n):
        return _cvxcore.IntVector_reserve(self, n)

    def capacity(self):
        return _cvxcore.IntVector_capacity(self)
    __swig_destroy__ = _cvxcore.delete_IntVector
    __del__ = lambda self: None
IntVector_swigregister = _cvxcore.IntVector_swigregister
IntVector_swigregister(IntVector)

class DoubleVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DoubleVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DoubleVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _cvxcore.DoubleVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __bool__(self):
        return _cvxcore.DoubleVector___nonzero__(self)

    def __bool__(self):
        return _cvxcore.DoubleVector___bool__(self)

    def __len__(self):
        return _cvxcore.DoubleVector___len__(self)

    def __getslice__(self, i, j):
        return _cvxcore.DoubleVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _cvxcore.DoubleVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _cvxcore.DoubleVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _cvxcore.DoubleVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _cvxcore.DoubleVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _cvxcore.DoubleVector___setitem__(self, *args)

    def pop(self):
        return _cvxcore.DoubleVector_pop(self)

    def append(self, x):
        return _cvxcore.DoubleVector_append(self, x)

    def empty(self):
        return _cvxcore.DoubleVector_empty(self)

    def size(self):
        return _cvxcore.DoubleVector_size(self)

    def swap(self, v):
        return _cvxcore.DoubleVector_swap(self, v)

    def begin(self):
        return _cvxcore.DoubleVector_begin(self)

    def end(self):
        return _cvxcore.DoubleVector_end(self)

    def rbegin(self):
        return _cvxcore.DoubleVector_rbegin(self)

    def rend(self):
        return _cvxcore.DoubleVector_rend(self)

    def clear(self):
        return _cvxcore.DoubleVector_clear(self)

    def get_allocator(self):
        return _cvxcore.DoubleVector_get_allocator(self)

    def pop_back(self):
        return _cvxcore.DoubleVector_pop_back(self)

    def erase(self, *args):
        return _cvxcore.DoubleVector_erase(self, *args)

    def __init__(self, *args):
        this = _cvxcore.new_DoubleVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _cvxcore.DoubleVector_push_back(self, x)

    def front(self):
        return _cvxcore.DoubleVector_front(self)

    def back(self):
        return _cvxcore.DoubleVector_back(self)

    def assign(self, n, x):
        return _cvxcore.DoubleVector_assign(self, n, x)

    def resize(self, *args):
        return _cvxcore.DoubleVector_resize(self, *args)

    def insert(self, *args):
        return _cvxcore.DoubleVector_insert(self, *args)

    def reserve(self, n):
        return _cvxcore.DoubleVector_reserve(self, n)

    def capacity(self):
        return _cvxcore.DoubleVector_capacity(self)
    __swig_destroy__ = _cvxcore.delete_DoubleVector
    __del__ = lambda self: None
DoubleVector_swigregister = _cvxcore.DoubleVector_swigregister
DoubleVector_swigregister(DoubleVector)

class IntVector2D(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, IntVector2D, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, IntVector2D, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _cvxcore.IntVector2D_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __bool__(self):
        return _cvxcore.IntVector2D___nonzero__(self)

    def __bool__(self):
        return _cvxcore.IntVector2D___bool__(self)

    def __len__(self):
        return _cvxcore.IntVector2D___len__(self)

    def __getslice__(self, i, j):
        return _cvxcore.IntVector2D___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _cvxcore.IntVector2D___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _cvxcore.IntVector2D___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _cvxcore.IntVector2D___delitem__(self, *args)

    def __getitem__(self, *args):
        return _cvxcore.IntVector2D___getitem__(self, *args)

    def __setitem__(self, *args):
        return _cvxcore.IntVector2D___setitem__(self, *args)

    def pop(self):
        return _cvxcore.IntVector2D_pop(self)

    def append(self, x):
        return _cvxcore.IntVector2D_append(self, x)

    def empty(self):
        return _cvxcore.IntVector2D_empty(self)

    def size(self):
        return _cvxcore.IntVector2D_size(self)

    def swap(self, v):
        return _cvxcore.IntVector2D_swap(self, v)

    def begin(self):
        return _cvxcore.IntVector2D_begin(self)

    def end(self):
        return _cvxcore.IntVector2D_end(self)

    def rbegin(self):
        return _cvxcore.IntVector2D_rbegin(self)

    def rend(self):
        return _cvxcore.IntVector2D_rend(self)

    def clear(self):
        return _cvxcore.IntVector2D_clear(self)

    def get_allocator(self):
        return _cvxcore.IntVector2D_get_allocator(self)

    def pop_back(self):
        return _cvxcore.IntVector2D_pop_back(self)

    def erase(self, *args):
        return _cvxcore.IntVector2D_erase(self, *args)

    def __init__(self, *args):
        this = _cvxcore.new_IntVector2D(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _cvxcore.IntVector2D_push_back(self, x)

    def front(self):
        return _cvxcore.IntVector2D_front(self)

    def back(self):
        return _cvxcore.IntVector2D_back(self)

    def assign(self, n, x):
        return _cvxcore.IntVector2D_assign(self, n, x)

    def resize(self, *args):
        return _cvxcore.IntVector2D_resize(self, *args)

    def insert(self, *args):
        return _cvxcore.IntVector2D_insert(self, *args)

    def reserve(self, n):
        return _cvxcore.IntVector2D_reserve(self, n)

    def capacity(self):
        return _cvxcore.IntVector2D_capacity(self)
    __swig_destroy__ = _cvxcore.delete_IntVector2D
    __del__ = lambda self: None
IntVector2D_swigregister = _cvxcore.IntVector2D_swigregister
IntVector2D_swigregister(IntVector2D)

class DoubleVector2D(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DoubleVector2D, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DoubleVector2D, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _cvxcore.DoubleVector2D_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __bool__(self):
        return _cvxcore.DoubleVector2D___nonzero__(self)

    def __bool__(self):
        return _cvxcore.DoubleVector2D___bool__(self)

    def __len__(self):
        return _cvxcore.DoubleVector2D___len__(self)

    def __getslice__(self, i, j):
        return _cvxcore.DoubleVector2D___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _cvxcore.DoubleVector2D___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _cvxcore.DoubleVector2D___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _cvxcore.DoubleVector2D___delitem__(self, *args)

    def __getitem__(self, *args):
        return _cvxcore.DoubleVector2D___getitem__(self, *args)

    def __setitem__(self, *args):
        return _cvxcore.DoubleVector2D___setitem__(self, *args)

    def pop(self):
        return _cvxcore.DoubleVector2D_pop(self)

    def append(self, x):
        return _cvxcore.DoubleVector2D_append(self, x)

    def empty(self):
        return _cvxcore.DoubleVector2D_empty(self)

    def size(self):
        return _cvxcore.DoubleVector2D_size(self)

    def swap(self, v):
        return _cvxcore.DoubleVector2D_swap(self, v)

    def begin(self):
        return _cvxcore.DoubleVector2D_begin(self)

    def end(self):
        return _cvxcore.DoubleVector2D_end(self)

    def rbegin(self):
        return _cvxcore.DoubleVector2D_rbegin(self)

    def rend(self):
        return _cvxcore.DoubleVector2D_rend(self)

    def clear(self):
        return _cvxcore.DoubleVector2D_clear(self)

    def get_allocator(self):
        return _cvxcore.DoubleVector2D_get_allocator(self)

    def pop_back(self):
        return _cvxcore.DoubleVector2D_pop_back(self)

    def erase(self, *args):
        return _cvxcore.DoubleVector2D_erase(self, *args)

    def __init__(self, *args):
        this = _cvxcore.new_DoubleVector2D(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _cvxcore.DoubleVector2D_push_back(self, x)

    def front(self):
        return _cvxcore.DoubleVector2D_front(self)

    def back(self):
        return _cvxcore.DoubleVector2D_back(self)

    def assign(self, n, x):
        return _cvxcore.DoubleVector2D_assign(self, n, x)

    def resize(self, *args):
        return _cvxcore.DoubleVector2D_resize(self, *args)

    def insert(self, *args):
        return _cvxcore.DoubleVector2D_insert(self, *args)

    def reserve(self, n):
        return _cvxcore.DoubleVector2D_reserve(self, n)

    def capacity(self):
        return _cvxcore.DoubleVector2D_capacity(self)
    __swig_destroy__ = _cvxcore.delete_DoubleVector2D
    __del__ = lambda self: None
DoubleVector2D_swigregister = _cvxcore.DoubleVector2D_swigregister
DoubleVector2D_swigregister(DoubleVector2D)

class IntIntMap(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, IntIntMap, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, IntIntMap, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _cvxcore.IntIntMap_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __bool__(self):
        return _cvxcore.IntIntMap___nonzero__(self)

    def __bool__(self):
        return _cvxcore.IntIntMap___bool__(self)

    def __len__(self):
        return _cvxcore.IntIntMap___len__(self)
    def __iter__(self):
        return self.key_iterator()
    def iterkeys(self):
        return self.key_iterator()
    def itervalues(self):
        return self.value_iterator()
    def iteritems(self):
        return self.iterator()

    def __getitem__(self, key):
        return _cvxcore.IntIntMap___getitem__(self, key)

    def __delitem__(self, key):
        return _cvxcore.IntIntMap___delitem__(self, key)

    def has_key(self, key):
        return _cvxcore.IntIntMap_has_key(self, key)

    def keys(self):
        return _cvxcore.IntIntMap_keys(self)

    def values(self):
        return _cvxcore.IntIntMap_values(self)

    def items(self):
        return _cvxcore.IntIntMap_items(self)

    def __contains__(self, key):
        return _cvxcore.IntIntMap___contains__(self, key)

    def key_iterator(self):
        return _cvxcore.IntIntMap_key_iterator(self)

    def value_iterator(self):
        return _cvxcore.IntIntMap_value_iterator(self)

    def __setitem__(self, *args):
        return _cvxcore.IntIntMap___setitem__(self, *args)

    def asdict(self):
        return _cvxcore.IntIntMap_asdict(self)

    def __init__(self, *args):
        this = _cvxcore.new_IntIntMap(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def empty(self):
        return _cvxcore.IntIntMap_empty(self)

    def size(self):
        return _cvxcore.IntIntMap_size(self)

    def swap(self, v):
        return _cvxcore.IntIntMap_swap(self, v)

    def begin(self):
        return _cvxcore.IntIntMap_begin(self)

    def end(self):
        return _cvxcore.IntIntMap_end(self)

    def rbegin(self):
        return _cvxcore.IntIntMap_rbegin(self)

    def rend(self):
        return _cvxcore.IntIntMap_rend(self)

    def clear(self):
        return _cvxcore.IntIntMap_clear(self)

    def get_allocator(self):
        return _cvxcore.IntIntMap_get_allocator(self)

    def count(self, x):
        return _cvxcore.IntIntMap_count(self, x)

    def erase(self, *args):
        return _cvxcore.IntIntMap_erase(self, *args)

    def find(self, x):
        return _cvxcore.IntIntMap_find(self, x)

    def lower_bound(self, x):
        return _cvxcore.IntIntMap_lower_bound(self, x)

    def upper_bound(self, x):
        return _cvxcore.IntIntMap_upper_bound(self, x)
    __swig_destroy__ = _cvxcore.delete_IntIntMap
    __del__ = lambda self: None
IntIntMap_swigregister = _cvxcore.IntIntMap_swigregister
IntIntMap_swigregister(IntIntMap)

class LinOpVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, LinOpVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, LinOpVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _cvxcore.LinOpVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __bool__(self):
        return _cvxcore.LinOpVector___nonzero__(self)

    def __bool__(self):
        return _cvxcore.LinOpVector___bool__(self)

    def __len__(self):
        return _cvxcore.LinOpVector___len__(self)

    def __getslice__(self, i, j):
        return _cvxcore.LinOpVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _cvxcore.LinOpVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _cvxcore.LinOpVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _cvxcore.LinOpVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _cvxcore.LinOpVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _cvxcore.LinOpVector___setitem__(self, *args)

    def pop(self):
        return _cvxcore.LinOpVector_pop(self)

    def append(self, x):
        return _cvxcore.LinOpVector_append(self, x)

    def empty(self):
        return _cvxcore.LinOpVector_empty(self)

    def size(self):
        return _cvxcore.LinOpVector_size(self)

    def swap(self, v):
        return _cvxcore.LinOpVector_swap(self, v)

    def begin(self):
        return _cvxcore.LinOpVector_begin(self)

    def end(self):
        return _cvxcore.LinOpVector_end(self)

    def rbegin(self):
        return _cvxcore.LinOpVector_rbegin(self)

    def rend(self):
        return _cvxcore.LinOpVector_rend(self)

    def clear(self):
        return _cvxcore.LinOpVector_clear(self)

    def get_allocator(self):
        return _cvxcore.LinOpVector_get_allocator(self)

    def pop_back(self):
        return _cvxcore.LinOpVector_pop_back(self)

    def erase(self, *args):
        return _cvxcore.LinOpVector_erase(self, *args)

    def __init__(self, *args):
        this = _cvxcore.new_LinOpVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _cvxcore.LinOpVector_push_back(self, x)

    def front(self):
        return _cvxcore.LinOpVector_front(self)

    def back(self):
        return _cvxcore.LinOpVector_back(self)

    def assign(self, n, x):
        return _cvxcore.LinOpVector_assign(self, n, x)

    def resize(self, *args):
        return _cvxcore.LinOpVector_resize(self, *args)

    def insert(self, *args):
        return _cvxcore.LinOpVector_insert(self, *args)

    def reserve(self, n):
        return _cvxcore.LinOpVector_reserve(self, n)

    def capacity(self):
        return _cvxcore.LinOpVector_capacity(self)
    __swig_destroy__ = _cvxcore.delete_LinOpVector
    __del__ = lambda self: None
LinOpVector_swigregister = _cvxcore.LinOpVector_swigregister
LinOpVector_swigregister(LinOpVector)


def build_matrix(*args):
    return _cvxcore.build_matrix(*args)
build_matrix = _cvxcore.build_matrix
# This file is compatible with both classic and new-style classes.


